# -*- coding: ascii -*-
# $Id$
# $Source$
# $Author$
# $HeadURL$
# $Revision$
# $Date$
'''
    roleplay:

        Implements Roles for Python.

        See the Traits paper[1], Perl 6 Synopsis 12[2] or Perl 6 Apocalypse 12[3]
        for more information.


    author: Ask Solem <askh@opera.com>
    Copyright (c) Ask Solem. Released under the Modified BSD-license, see the
    LICENSE file for a full copy of the license.
     
'''


__version__     = '0.11'
__author__      = 'Ask Solem <askh@opera.com'
__authority__   = 'pypi:ASK'


from roleplay.keyword import *


# Local Variables:
#   mode: cpython
#   cpython-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab tabstop=4 shiftwidth=4 shiftround



# -*- coding: ascii -*-
# $Id$
# $Source$
# $Author$
# $HeadURL$
# $Revision$
# $Date$
'''
    roleplay.inheritable:

        Base class with method versions of does() and has_role().

        You don't really have to inherit from this class, you can just as well
        import the function versions from the roleplay top-level module.

            from roleplay import does, has_role

        However, if you already have a common base class for all your objects, or you
        maintain a metaclass, this module could be the thing for you.

    author: Ask Solem <askh@opera.com>
    Copyright (c) Ask Solem. Released under the Modified BSD-license, see the
    LICENSE file for a full copy of the license.

'''

__version__     = '0.11'
__author__      = 'Ask Solem <askh@opera.com'
__authority__   = 'pypi:ASK'

import roleplay.keyword

class RoleObject(object):
    '''
        Base class with method versions of does() and has_role().

        Example usage:

            from some.module import SomeRole

            class MyClass(RoleObject):
                
                def __init__(self):
                    self.has_role(SomeRole)

                def do_something(self):

                    # Test if we support the SomeRole role, if so we get    
                    # extra functonality.
                    if self.does('SomeRole'):
                        self.some_method_SomeRole_defines()

    '''

    def has_role(self, role, **kwargs):
        '''
            Method version of roleplay.has_role

            %{role} must be class, not object instance.

            Example:

                self.has_role(RoleClass)
        '''
        roleplay.keyword.has_role(self, role, **kwargs)

    def does(self, role):
        '''
            Method version of roleplay.does

            Returns True if this class supports the role.

            %(role) can be instance, class or string.

            Example:

                has_comment_support = self.does('Comments')
        '''
        roleplay.keyword.does(self, role)


# Local Variables:
#   mode: cpython
#   cpython-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab tabstop=4 shiftwidth=4 shiftround
    







# -*- coding: ascii -*-
# $Id$
# $Source$
# $Author$
# $HeadURL$
# $Revision$
# $Date$
'''
    roleplay.keyword:

        Apply and test for roles.

    author: Ask Solem <askh@opera.com>
    Copyright (c) Ask Solem. Released under the Modified BSD-license, see the
    LICENSE file for a full copy of the license.
'''

__version__     = '0.11'
__author__      = 'Ask Solem <askh@opera.com'
__authority__   = 'pypi:ASK'

from roleplay.meta import MetaRole
from roleplay.role import Role

DEBUG = 0
DOES_FRAME_STEP_MAX = 2

# this should be list(dir(Role)) when the Role baseclass is finished

class DoesOutsideClass(Exception):
    ''' does() must be used in a class definiton '''

def has_role(instance, role, **kwargs):
    '''
        Applies the role %{role} to your class.

        %{instance} can be either instance or class.
        %{role} must be class, not object instance.

        Example:

        has_role(self, RoleClass)


    '''
    role(instance, **kwargs).__buildrole__()

def does(instance, role):
    '''
        Returns True if %{instance} supports the role %(role).

        %(class) can be instance, class or string.
        %(role) can be instance, class or string.

        Example:

        has_comment_support = does(self, 'Comments')
    '''
    meta = MetaRole()
    return meta.__does__(instance, role)

''' XXX Not finished 
def doesx(role, **kwargs):
    caller = ""
    try:
        frame  = sys._getframe(1)
    except ValueError:
        raise DesOutsideClass;
    
    frameinfo   = inspect.getframeinfo(frame, context=1)
    module_file = frameinfo[0]
    cls         = frameinfo[2] 
    
    module_name, module_ext = os.path.splitext(module_file)
   
    module = sys.modules.get(module_name) 
    
    print "module: [%s], module_file: [%s], module_name: [%s], ext: [%s], class: [%s]" % (
        str(module), module_file, module_name, module_ext, cls
    )
'''


# Local Variables:
#   mode: cpython
#   cpython-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab tabstop=4 shiftwidth=4 shiftround



# -*- coding: ascii -*-
# $Id$
# $Source$
# $Author$
# $HeadURL$
# $Revision$
# $Date$
'''
    roleplay.meta:

        This module handles internal role metadata and the introspection
        necessary to install roles and mixin classes.

    author: Ask Solem <askh@opera.com>
    Copyright (c) Ask Solem. Released under the Modified BSD-license, see the
    LICENSE file for a full copy of the license.
'''

__version__     = '0.11'
__author__      = 'Ask Solem <askh@opera.com'
__authority__   = 'pypi:ASK'

import os
import sys
import types
import inspect
from   itertools import ifilter
from   inspect   import isbuiltin

DEBUG = 0

roles_defines_attrs = "check_requires for_class role_args meta".split()

# ##### EXCEPTIONS

class AttributeIsPrototypeError(Exception):
    ''' The attribute is only a prototype and is not accessible '''

# #### DEBUGGING
def _announce_symbol_exported(from_class, symbol, into):
    ''' MetaRole._mixin() uses this to print whenever a symbol was exported
        in DEBUG mode
    '''
    print (">>> Role %(from_class)s: "
           "Exporting symbol [%(symbol)s] into %(into)s" % locals())

def get_class_name(cls):
    '''
        Given a class, instance or string this always
        tries to return the class name
    '''
    name = ""
    if isinstance(cls, object):
        if hasattr(cls, '__name__'):
            name = cls.__name__
        else:   # is instance
            name = cls.__class__.__name__
    else:
        name = cls
    return name

class MetaRoleRegistry(object):
    '''
        Keeps track of classes and their roles.
    '''
    _shared_state   = { }
    _role_registry  = { }
    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls, *args, **kwargs)
        obj.__dict__ = cls._shared_state
        return obj

    def get_roles_for_class(self, cls):
        name = get_class_name(cls)
        if name not in self._role_registry:
            self._role_registry[name] = { }
        return self._role_registry[name]
            
class MetaRole(object):
    ''' -------------------------------------------  --- ---- -   - -  - -
        Roles metadata.

        Probably only used by the Role class itself. If you want to
        subclass the role system, it's probably better to subclass Role
        rather than MetaRole.
    ---------------------------------------------------------------------- '''

    # ### Prototype attributes
    # only here so they can be skipped when mixing.
    for_class  = None
    role_args  = None
    meta       = None

    # ### Registry composite
    role_registry = MetaRoleRegistry()
        
    
    def __buildrole__(self, role_class):
        ''' This is the step that installs the role into the class. '''
        self._mixin(role_class, role_class.for_class)
        self.__addrole__(role_class.for_class, role_class)

    def __does__(self, cls, role):
        ''' private method doing the role lookup '''
        roles_for = self.role_registry.get_roles_for_class(cls)
        role_name = get_class_name(role)
        if role_name in roles_for:
            return True
        else:
            return False

    def __addrole__(self, cls, role):
        ''' private method updating the role registry '''
        roles_for = self.role_registry.get_roles_for_class(cls)
        role_name = get_class_name(role)
        roles_for[role_name] = True

    def init_attributes(self, cls, **kwargs):
        '''
            Initialize multiple attributes at once, with default values.
            Usually used within the __init__ method of a class.

            Example usage:

                class MyRole(Role):
                    meta = MetaRole()

                    def __init__(self, for_class):
                        meta.init_attributes(self,
                            for_class=for_class,
                        )

            This will initialize the attributes for_class
            with the default values passed when constructing the instance.
        '''
        [setattr(cls, name, value) for name, value in kwargs.iteritems()]

    def apply_attribute(self, for_class, attribute_name, attribute_value):
        if hasattr(for_class, attribute_name):
            return
        setattr(for_class, attribute_name, attribute_value)


    def __getattribute__(self, attribute_name):
        ''' Same as the usual __getattr__ except it will raise an
            AttributeIsPrototypeError when accessing any attributes
            that just defines the instance of a real Role.
        '''
        if attribute_name in roles_defines_attrs:
            raise AttributeIsPrototypeError
        return super(MetaRole, self).__getattribute__(attribute_name)

    def _get_normalized_symtable(self, from_object):
        '''
            Get a list of normalized symbols from an object.

            This is like dir(from_object), except it will not contain
            symbols from builtin methods, the type metaclass,
            or the MetaRole class.
        '''
        results = []
        for symbol in dir(from_object):
            if symbol in skip_sym:
                continue
            value = getattr(from_object, symbol)
            if isinstance(value, types.MethodType):
                results.append((symbol, value))
        return results

    def _mixin(self, from_class, into_other):
        '''
            Installs all attributes/methods in the role into the do'ing
            class that the class has not defined.

            It will not install builtin methods or attributes/methods from
            object (the type metaclass), or the MetaRole class.
        '''
        others_symtable = set( dir(into_other) )
        
        for symbol, symbol_ref in self._get_normalized_symtable(from_class):

            if (not isbuiltin(symbol_ref) and symbol not in others_symtable):
                setattr(into_other, symbol, symbol_ref)

                if DEBUG: _announce_symbol_exported(from_class, symbol, into_other)


'''
set: skip_sym
    Used by MetaRole._get_normalized_symtable_for() as a list of
    attributes/methods we don't export when doing a mixin.
'''  
skip_sym = set( dir(MetaRole) )
skip_sym.add('__metaclass__')

# Local Variables:
#   mode: cpython
#   cpython-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab tabstop=4 shiftwidth=4 shiftround



# -*- coding: ascii -*-
# $Id$
# $Source$
# $Author$
# $HeadURL$
# $Revision$
# $Date$
'''
    pyrole.role:

        Module for creating roles.

    author: Ask Solem <askh@opera.com>
    Copyright (c) Ask Solem. Released under the Modified BSD-license, see the
    LICENSE file for a full copy of the license.
'''

__version__     = '0.11'
__author__      = 'Ask Solem <askh@opera.com'
__authority__   = 'pypi:ASK'

from roleplay.meta import MetaRole

class ClassDoesNotFulfillRequirement(Exception):
    ''' Tried to apply role to the class, but the class didn't fulfill
        all requirements of the role. (Probably missing method/attribute)
    '''

class Role(object):
    ''' -------------------------------------------  --- ---- -   - -  - -
        Base class for Roles.

        New roles inherits from this.
    ---------------------------------------------------------------------- '''
    meta = MetaRole()

    def __init__(self, instance, **kwargs):
        self.meta.init_attributes(self, for_class=instance, role_args=kwargs)

    def apply_attribute(self, attr_name, attr_value):
        for_class = self.for_class
        self.meta.apply_attribute(for_class, attr_name, attr_value)

    def __buildrole__(self):
        ''' 
            Builds the role using the MetaRole class.
            If the role defines the '__requires__' attribute it
            also checks if the user class adheres to the
            required interface. (See 'check_requires' for more info)

        '''
       
        # Since we only do it by instance now we have to add
        # the role for every constructed instance :( 
        '''    
        # Just return if the class already does this role.
        for_class = self.for_class
        if self.meta.__does__(for_class, self):
            return
        '''

        if hasattr(self, '__requires__'):
            self.check_requires()

        self.meta.__buildrole__(self)

    def check_requires(self):
        '''
            The role can have a 'requires' attribute containing a list of
            attribute-names that the user class has to implement to do the
            role. This method will check that all requirements are fulfilled.
        '''
        for_class = self.for_class

        for requirement in self.__requires__:
            if not hasattr(for_class, requirement):
                raise ClassDoesNotFulfillRequirement(requirement)


# Local Variables:
#   mode: cpython
#   cpython-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab tabstop=4 shiftwidth=4 shiftround



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import codecs

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages


setup(
    name='roleplay',
    version='0.11',
    description='Python does Roles',
    author='Ask Solem',
    author_email='askh@opera.com',
    packages=find_packages(exclude=['ez_setup', 'examples']),
    url='http://pypi.python.org/pypi/roleplay/',
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    long_description=codecs.open('README.txt', "r", "utf-8").read(),
)
